from abc import ABC, abstractmethod
import numpy as np
# the following functions are taken from Ben Southgate:
# https://bsouthga.dev/posts/colour-gradients-with-python


def hex_to_RGB(hex):
    """ "#FFFFFF" -> [255,255,255]"""
    # Pass 16 to the integer function for change of base
    return [int(hex[i : i + 2], 16) for i in range(1, 6, 2)]


def RGB_to_hex(RGB):
    """[255,255,255] -> "#FFFFFF" """
    # Components need to be integers for hex to make sense
    RGB = [int(x) for x in RGB]
    return "#" + "".join(
        ["0{0:x}".format(v) if v < 16 else "{0:x}".format(v) for v in RGB]
    )


def colour_dict(gradient):
    """Takes in a list of RGB sub-lists and returns dictionary of
    colours in RGB and hex form for use in a graphing function
    defined later on."""
    return {
        "hex": [RGB_to_hex(RGB) for RGB in gradient],
        "r": [RGB[0] for RGB in gradient],
        "g": [RGB[1] for RGB in gradient],
        "b": [RGB[2] for RGB in gradient],
    }


def linear_gradient(start_hex, finish_hex="#FFFFFF", n=10):
    """returns a gradient list of (n) colours between
    two hex colours. start_hex and finish_hex
    should be the full six-digit colour string,
    inlcuding the number sign ("#FFFFFF")"""
    # Starting and ending colours in RGB form
    s = hex_to_RGB(start_hex)
    f = hex_to_RGB(finish_hex)
    # Initilize a list of the output colours with the starting colour
    RGB_list = [s]
    # Calcuate a colour at each evenly spaced value of t from 1 to n
    for t in range(0, n):
        # Interpolate RGB vector for colour at the current value of t
        curr_vector = [
            int(s[j] + (float(t) / (n - 1)) * (f[j] - s[j])) for j in range(3)
        ]
        # Add it to our list of output colours
        RGB_list.append(curr_vector)

    return colour_dict(RGB_list)


def polylinear_gradient(colours, n):
    """returns a list of colours forming linear gradients between
    all sequential pairs of colours. "n" specifies the total
    number of desired output colours"""
    # The number of colours per individual linear gradient
    n_out = int(float(n) / (len(colours) - 1))
    # returns dictionary defined by colour_dict()
    gradient_dict = linear_gradient(colours[0], colours[1], n_out)

    if len(colours) > 1:
        for col in range(1, len(colours) - 1):
            next = linear_gradient(colours[col], colours[col + 1], n_out)
            for k in ("hex", "r", "g", "b"):
                # Exclude first point to avoid duplicates
                gradient_dict[k] += next[k][1:]

    return gradient_dict


class ColourMap(ABC):
    @abstractmethod
    def __call__(self, v: float): ...


class LinearGradientColourMap(ColourMap):
    def __init__(
        self,
        colours: list[str] | None = ["#ff0000", "#ffffff", "#0000ff"],
        min_value: float | None = 0,
        max_value: float | None = 1,
        bins: int = 100,
    ):
        self.colours = polylinear_gradient(colours, bins)
        self.min, self.max = min_value, max_value

    def __call__(self, v: float):
        v = max(0, int((v - self.min) / (self.max - self.min) * 100) - 1)
        if v >= len(self.colours["hex"]):
            breakpoint()
        return self.colours["hex"][v]


class RandomColourMap(ColourMap):
    def __init__(self, random_state: int | list[int] | None = [2, 3, 4, 5, 6]):
        self.rgen = np.random.default_rng(random_state)

    def __call__(self, v: float):
        return RGB_to_hex([x * 255 for x in self.rgen.random(3)])
